from abc import ABCMeta, abstractmethod

from guniflask.security.authentication import Authentication


class AuthenticationManager(metaclass=ABCMeta):
    @abstractmethod
    def authenticate(self, authentication: Authentication) -> Authentication:
        pass  # pragma: no cover



from guniflask.utils.rule import make_ignore_rule_for_field, make_only_rule_for_field, _make_rule_set


def test_make_set():
    assert _make_rule_set('a') == {'a'}
    assert _make_rule_set('a,b') == {'a', 'b'}
    assert _make_rule_set({'a'}) == {'a'}
    assert _make_rule_set(['a', 'b']) == {'a', 'b'}


def test_make_ignore_rule():
    assert make_ignore_rule_for_field('a.b') == {'a.b'}


def test_make_only_rule():
    assert make_only_rule_for_field('a.b') == {'a', 'a.b'}



from typing import Optional

from guniflask.context.annotation import autowired
from guniflask.context.bean_context import BeanContext, BeanContextAware
from guniflask.security.authentication import Authentication
from guniflask.security.authentication_manager import AuthenticationManager
from guniflask.security_config.authentication_config import AuthenticationConfiguration
from guniflask.security_config.authentication_manager_builder import AuthenticationManagerBuilder
from guniflask.security_config.http_security import HttpSecurity
from guniflask.security_config.security_configurer import SecurityConfigurer
from guniflask.security_config.web_security import WebSecurity


class WebSecurityConfigurer(SecurityConfigurer, BeanContextAware):
    def __init__(self):
        super().__init__()
        self._authentication_configuration: Optional[AuthenticationConfiguration] = None
        self._authentication_builder = AuthenticationManagerBuilder()
        self._local_authentication_builder = AuthenticationManagerBuilder()
        self._enable_local_authentication = False
        self._authentication_manager: Optional[AuthenticationManager] = None
        self._authentication_manager_initialized = False
        self._http: Optional[HttpSecurity] = None
        self._context: Optional[BeanContext] = None

    def init(self, web_security: WebSecurity):
        http = self._get_http()
        web_security.add_security_builder(http)

    def configure(self, web_security: WebSecurity):
        pass

    def configure_http(self, http: HttpSecurity):
        pass

    def configure_authentication(self, authentication_builder: AuthenticationManagerBuilder):
        pass

    def authentication_manager_bean(self) -> AuthenticationManager:
        return AuthenticationManagerDelegator(self._authentication_builder)

    def set_bean_context(self, bean_context: BeanContext):
        self._context = bean_context

    @autowired
    def set_authentication_config(self, authentication_configuration: AuthenticationConfiguration):
        self._authentication_configuration = authentication_configuration

    def _get_http(self) -> HttpSecurity:
        if self._http:
            return self._http
        authentication_manager = self._get_authentication_manager()
        self._authentication_builder.with_parent_authentication_manager(authentication_manager)
        shared_objects = self._create_shared_objects()

        self._http = HttpSecurity(self._authentication_builder, shared_objects=shared_objects)
        self.configure_http(self._http)
        return self._http

    def _get_authentication_manager(self) -> AuthenticationManager:
        if not self._authentication_manager_initialized:
            self.configure_authentication(self._local_authentication_builder)
            if self._enable_local_authentication:
                self._authentication_manager = self._local_authentication_builder.build()
            else:
                self._authentication_manager = self._authentication_configuration.authentication_manager
            self._authentication_manager_initialized = True
        return self._authentication_manager

    def _create_shared_objects(self):
        shared_objects = {}
        for k, v in self._local_authentication_builder.get_shared_objects():
            shared_objects[k] = v
        shared_objects[BeanContext] = self._context
        return shared_objects


class AuthenticationManagerDelegator(AuthenticationManager):
    def __init__(self, delegate_builder: AuthenticationManagerBuilder):
        self._delegate_builder = delegate_builder
        self._delegate: AuthenticationManager = None

    def authenticate(self, authentication: Authentication) -> Authentication:
        if self._delegate is None:
            self._delegate = self._delegate_builder.object
            self._delegate_builder = None
        return self._delegate.authenticate(authentication)



import base64
import os
from json import JSONEncoder as _JSONEncoder

import jwt


class JsonEncoder(_JSONEncoder):
    def default(self, o):
        if isinstance(o, set):
            return list(o)
        return super().default(o)


class JwtHelper:
    @staticmethod
    def generate_jwt_secret(n=32):
        return base64.b64encode(os.urandom(n), altchars=b'-_').decode().replace('=', '')

    @staticmethod
    def decode_jwt(token, key, algorithms=None, **kwargs):
        if algorithms is None:
            header = jwt.get_unverified_header(token)
            algorithms = [header['alg']]

        return jwt.decode(token, key=key, algorithms=algorithms, **kwargs)

    @staticmethod
    def encode_jwt(payload, key, algorithm=None, **kwargs):
        token = jwt.encode(payload, key, algorithm=algorithm or 'HS256', json_encoder=JsonEncoder, **kwargs)
        if isinstance(token, bytes):
            token = token.decode('utf-8')
        return token



from guniflask.scheduling import scheduled, async_run
from guniflask.web import blueprint, post_route, get_route


@blueprint
class SchedulerController:

    def __init__(self):
        self.scheduled = False
        self.async_result = None

    @scheduled(interval=1)
    def schedule_task(self):
        self.scheduled = True

    @get_route('/scheduled')
    def get_scheduled(self):
        return {'result': self.scheduled}

    @post_route('/async-add')
    def async_add(self, x: float, y: float):
        self.do_async_add(x, y)
        return {}

    @async_run
    def do_async_add(self, x: float, y: float):
        self.async_result = x + y

    @get_route('/async-add')
    def get_async_add_result(self):
        return {'result': self.async_result}



from flask_cors.core import parse_resources, serialize_options, DEFAULT_OPTIONS
from flask_cors.extension import make_after_request_function

from guniflask.web.request_filter import RequestFilter


class CorsFilter(RequestFilter):

    def __init__(self, **kwargs):
        options = self._get_cors_options(kwargs)
        resources = parse_resources(options.get('resources'))
        resources = [
            (pattern, self._get_cors_options(options, opts))
            for (pattern, opts) in resources
        ]
        self._after_request = make_after_request_function(resources)

    def _get_cors_options(self, *dicts):
        options = dict(DEFAULT_OPTIONS)
        for d in dicts:
            options.update(d)
        return serialize_options(options)

    def after_request(self, response):
        return self._after_request(response)



from abc import ABCMeta, abstractmethod


class InitializingBean(metaclass=ABCMeta):
    @abstractmethod
    def after_properties_set(self):
        pass  # pragma: no cover


class SmartInitializingSingleton(metaclass=ABCMeta):
    @abstractmethod
    def after_singletons_instantiated(self):
        pass  # pragma: no cover


class DisposableBean(metaclass=ABCMeta):
    @abstractmethod
    def destroy(self):
        pass  # pragma: no cover



class BeanDefinition:
    SCOPE_SINGLETON = 'singleton'
    SCOPE_PROTOTYPE = 'prototype'

    def __init__(self, source):
        self._source = source
        self.scope = self.SCOPE_SINGLETON
        self.factory_bean_name = None

    @property
    def source(self):
        return self._source

    def is_singleton(self) -> bool:
        return self.scope == self.SCOPE_SINGLETON

    def is_prototype(self) -> bool:
        return self.scope == self.SCOPE_PROTOTYPE



from contextlib import contextmanager

from guniflask.app.factory import create_app
from guniflask.test.env import set_test_env


@contextmanager
def app_test_client():
    set_test_env()
    app = create_app()
    with app.app_context():
        with app.test_client() as client:
            yield client
